import logging
import time

import ansible.constants as C
from ansible.runner import Runner

from dopy.manager import DoManager, DoError
from domain import Box
from factory import FactoryProvider, FactoryCli
from simple import SimpleProvider
from utils.provisioning import run_module, local_inventory, create_user
from utils.io import input_yes_no, input_value, input_path, xstr


class DigitalOceanCli(FactoryCli):
    def __init__(self):
        FactoryCli.__init__(self)
        self.prompt = '(Prudentia > DigitalOcean) '
        self.provider = DigitalOceanProvider()


class DigitalOceanProvider(FactoryProvider):
    NAME = 'digital-ocean'

    DEFAULT_IMAGE_ID = 9801950  # Ubuntu 14.04 x64
    DEFAULT_SIZE_ID = 63  # 1GB
    DEFAULT_REGION_ID = 5  # Amsterdam 2

    def __init__(self):
        super(DigitalOceanProvider, self).__init__(self.NAME, DOGeneral, DOExt)
        if not self.env.initialized:
            g = self._input_general_env_conf()
        else:
            g = self.env.general
        self.manager = DoManager(g.client_id, g.api_key)

    def _input_general_env_conf(self):
        print '\nThis is the first time you use the Digital Ocean provider, please supply your credentials:'
        client_id = input_value('client id')
        api_key = input_value('api key')
        do_general = DOGeneral(client_id, api_key)
        self.env.set_general(do_general)
        return do_general

    def register(self):
        try:
            name = None
            ip = None
            ext = DOExt()

            if input_yes_no('register an existing droplet'):
                droplet_id = input_value('droplet id')
                droplet_info = self.manager.show_droplet(droplet_id)
                ext.id = droplet_id
                name = droplet_info['name']
                created = droplet_info['created_at']
                print '\nFound droplet \'{0}\' (created {1})'.format(name, created)
                ip = xstr(droplet_info['ip_address'])
                print 'IP: %s' % ip
                ext.image = droplet_info['image_id']
                print 'Image: %s' % ext.image
                ext.size = droplet_info['size_id']
                print 'Size: %s' % ext.size
                ext.region = droplet_info['region_id']
                print 'Region: %s\n' % ext.region

            playbook = input_path('absolute playbook path')
            hostname = self.fetch_box_hostname(playbook)

            if not name:
                name = input_value('box name', self.suggest_name(hostname))

            user = input_value('remote user', C.active_user)

            if not ext.image:
                all_images = self.manager.all_images()
                print '\nAvailable images: \n%s' % self._print_object_id_name(all_images)
                ext.image = input_value('image', self.DEFAULT_IMAGE_ID)

            if not ext.size:
                all_sizes = self.manager.sizes()
                print '\nAvailable sizes: \n%s' % self._print_object_id_name(all_sizes)
                ext.size = input_value('size', self.DEFAULT_SIZE_ID)

            if not ext.region:
                all_regions = self.manager.all_regions()
                print '\nAvailable regions: \n%s' % self._print_object_id_name(all_regions)
                ext.region = input_value('region', self.DEFAULT_REGION_ID)

            if not ext.keys:
                all_keys = self.manager.all_ssh_keys()
                print '\nAvailable keys: \n%s' % self._print_object_id_name(all_keys)
                default_keys = ','.join([str(k['id']) for k in all_keys])
                ext.keys = input_value('keys', default_keys)

            box = Box(name, playbook, hostname, ip, user, extra=ext)
            self.add_box(box)
            print "\nBox %s added." % box
        except Exception as e:
            logging.exception('Box not added.')
            print '\nError: %s\n' % e

    def _print_object_id_name(self, objs):
        return '\n'.join([str(o['id']) + ' -> ' + o['name'] for o in objs])

    def _find_object_name(self, objs, id):
        return next(o for o in objs if o['id'] == id)['name']

    def reconfigure(self, previous_box):
        try:
            self.remove_box(previous_box)

            playbook = input_path('absolute playbook path', previous_box.playbook)
            hostname = self.fetch_box_hostname(playbook)
            ip = previous_box.ip
            user = input_value('remote user', previous_box.remote_user)

            if not previous_box.extra.id:
                ext = DOExt()
                ext.id = previous_box.extra.id
                all_images = self.manager.all_images()
                print '\nAvailable images: \n%s' % self._print_object_id_name(all_images)
                ext.image = input_value('image', previous_box.extra.image)

                all_sizes = self.manager.sizes()
                print '\nAvailable sizes: \n%s' % self._print_object_id_name(all_sizes)
                ext.size = input_value('size', previous_box.extra.size)

                all_keys = self.manager.all_ssh_keys()
                print '\nAvailable keys: \n%s' % self._print_object_id_name(all_keys)
                ext.keys = input_value('keys', previous_box.extra.keys)

                all_regions = self.manager.all_regions()
                print '\nAvailable regions: \n%s' % self._print_object_id_name(all_regions)
                ext.region = input_value('region', previous_box.extra.region)
            else:
                ext = previous_box.extra

            box = Box(previous_box.name, playbook, hostname, ip, user, extra=ext)
            self.add_box(box)
            print "\nBox %s reconfigured." % box
        except Exception as e:
            logging.exception('Box not reconfigured.')
            print '\nError: %s\n' % e

    def add_box(self, box):
        if not box.ip:
            self.create(box)
        SimpleProvider.add_box(self, box)

    def create(self, box):
        g = self.env.general
        e = box.extra
        if not e.id:
            print '\nCreating instance \'{0}\' ...'.format(box.name)
            success, result = run_module(Runner(
                transport='local',
                inventory=local_inventory(),
                remote_user='root',
                module_name='digital_ocean',
                module_args='state=present command=droplet client_id={0} api_key={1} '
                            'name={2} size_id={3} image_id={4} region_id={5} ssh_key_ids={6} wait_timeout=300'
                .format(g.client_id, g.api_key, box.name, e.size, e.image, e.region, e.keys))
            )

            droplet = result['droplet']
            box.extra.id = droplet['id']
            box.ip = droplet['ip_address']
            print 'Droplet created with id: {0} -> {1}\n'.format(box.extra.id, box.ip)
        else:
            info = self.manager.show_droplet(e.id)
            print 'Droplet {0} already exists - status: {1}.'.format(e.id, info['status'])
        create_user(box)

    def start(self, box):
        box_id = box.extra.id
        print 'Starting droplet %s ...' % box_id
        self.manager.power_on_droplet(box_id)
        self._wait_to_be_active(box_id)

    def stop(self, box):
        box_id = box.extra.id
        print 'Stopping droplet %s ...' % box_id
        self.manager.power_off_droplet(box_id)

    def destroy(self, box):
        if input_yes_no('destroy the droplet \'{0}\''.format(box.name)):
            box_id = box.extra.id
            box.ip = None
            box.extra.id = None
            print 'Destroying droplet %s ...' % box_id
            self.manager.destroy_droplet(box_id, scrub_data=True)

    def rebuild(self, box):
        e = box.extra
        print 'Rebuilding droplet %s ...' % e.id
        self.manager.rebuild_droplet(e.id, e.image)
        self._wait_to_be_active(e.id)
        create_user(box)

    def status(self, box):
        print self.manager.show_droplet(box.extra.id)['status']

    def _wait_to_be_active(self, droplet_id, wait_timeout=300):
        end_time = time.time() + wait_timeout
        while time.time() < end_time:
            print 'Waiting for droplet %s to be active ...' % droplet_id
            time.sleep(min(20, end_time - time.time()))

            droplet = self.manager.show_droplet(droplet_id)
            if droplet['status'] == 'active':
                droplet_ip_address = droplet['ip_address']
                if not droplet_ip_address:
                    raise DoError('No ip is found.', droplet_id)
                print '\nDroplet %s is now active with ip %s\n' % (droplet_id, droplet_ip_address)
                time.sleep(10)  # Wait for some network latency ...
                return droplet_ip_address
        raise DoError('Wait for droplet running timeout', droplet_id)


class DOGeneral(object):
    def __init__(self, client_id, api_key):
        self.client_id = client_id
        self.api_key = api_key

    def __repr__(self):
        return 'DOGeneral[client_id: %s, api_key: %s]' % (self.client_id, self.api_key)

    def to_json(self):
        return {'client_id': self.client_id, 'api_key': self.api_key}

    @staticmethod
    def from_json(json):
        return DOGeneral(json['client_id'], json['api_key'])


class DOExt(object):
    id = None
    image = None
    size = None
    keys = None
    region = None

    def __repr__(self):
        return 'DOExt[id: %s, image: %s, size: %s, keys: %s, region: %s]' % (
            self.id, self.image, self.size, self.keys, self.region)

    def to_json(self):
        return {'id': self.id, 'image': self.image, 'size': self.size, 'keys': self.keys, 'region': self.region}

    @staticmethod
    def from_json(json):
        e = DOExt()
        e.id = json['id']
        e.image = json['image']
        e.size = json['size']
        e.keys = json['keys']
        e.region = json['region']
        return e



import logging

import ansible.constants as C
from domain import Box
from simple import SimpleProvider, SimpleCli
from utils.io import input_value, input_path
from utils.provisioning import run_playbook, generate_inventory


class LocalCli(SimpleCli):
    def __init__(self):
        SimpleCli.__init__(self)
        self.prompt = '(Prudentia > Local) '
        self.provider = LocalProvider()


class LocalProvider(SimpleProvider):
    NAME = 'local'

    def __init__(self):
        super(LocalProvider, self).__init__(self.NAME)

    def register(self):
        try:
            playbook = input_path('absolute playbook path')
            hostname = self.fetch_box_hostname(playbook)
            name = input_value('box name', self.suggest_name(hostname))

            box = Box(name, playbook, hostname, '127.0.0.1')
            self.add_box(box)
            print "\nBox %s added." % box
        except Exception as e:
            logging.exception('Box not added.')
            print '\nError: %s\n' % e

    def reconfigure(self, previous_box):
        try:
            self.remove_box(previous_box)

            playbook = input_path('absolute playbook path', previous_box.playbook)
            hostname = self.fetch_box_hostname(playbook)

            box = Box(previous_box.name, playbook, hostname, '127.0.0.1')
            self.add_box(box)
            print "\nBox %s reconfigured." % box
        except Exception as e:
            logging.exception('Box not reconfigured.')
            print '\nError: %s\n' % e

    def provision(self, box, *tags):
        remote_user = C.DEFAULT_REMOTE_USER
        if box.remote_user:
            remote_user = box.remote_user

        remote_pwd = C.DEFAULT_REMOTE_PASS
        if not box.use_ssh_key():
            remote_pwd = box.remote_pwd

        transport = 'local'
        box.use_prudentia_lib = True

        only_tags = None
        if tags is not ():
            only_tags = tags

        self.provisioned = run_playbook(
            playbook_file=box.playbook,
            inventory=generate_inventory(box),
            remote_user=remote_user,
            remote_pass=remote_pwd,
            transport=transport,
            extra_vars=self.extra_vars,
            only_tags=only_tags
        )



import json
import os
from os import path
import sys

from utils.io import xstr


class Environment(object):
    DEFAULT_ENVS_PATH = path.join(os.environ['PRUDENTIA_USER_DIR'], 'envs')
    DEFAULT_ENV_FILE_NAME = 'boxes.json'

    def __init__(self, id_env, general_type=None, box_extra_type=None, envs_path=DEFAULT_ENVS_PATH,
                 file_name=DEFAULT_ENV_FILE_NAME):
        self.id_env = id_env
        env_path = path.join(envs_path, id_env)
        if not os.path.exists(env_path):
            print 'Initializing environment {0} ({1}) ...'.format(id_env, env_path)
            os.makedirs(env_path)
        self.file = path.join(env_path, file_name)
        self.general_type = general_type
        self.box_extra_type = box_extra_type
        self.general = None
        self.boxes = {}
        try:
            with open(self.file):
                self._load()
                self.initialized = True
        except IOError:
            self.initialized = False

    def set_general(self, general):
        self.general = general
        self._save()

    def add(self, box):
        if box.name not in self.boxes:
            self.boxes[box.name] = box
            self._save()
        else:
            raise ValueError("Box name must be unique: '{0}' already exists!".format(box.name))

    def get(self, box_name):
        return self.boxes.get(box_name)

    def remove(self, box):
        b = self.boxes.pop(box.name, None)
        self._save()
        return b

    def _load(self):
        f = None
        try:
            f = open(self.file, 'r')
            json_objects = json.load(f)
            if self.general_type:
                self.general = self.general_type.from_json(json_objects[0])
                json_boxes = json_objects[1]
            else:
                json_boxes = json_objects
            for jb in json_boxes:
                b = Box.from_json(jb, self.box_extra_type)
                self.boxes[b.name] = b
        except IOError, e:
            print e
        finally:
            if f:
                f.close()

    def _save(self):
        json_boxes = [b.to_json() for b in self.boxes.values()]
        f = None
        try:
            f = open(self.file, 'w')
            if self.general_type:
                json.dump([self.general.to_json(), json_boxes], f)
            else:
                json.dump(json_boxes, f)
        except IOError, e:
            print e
        finally:
            if f:
                f.close()


class Box(object):
    def __init__(self, name, playbook, hostname, ip, remote_user=None, remote_pwd=None, extra=None, use_prudentia_lib=False):
        self.name = name
        self.playbook = playbook
        self.hostname = hostname
        self.ip = ip
        self.remote_user = remote_user
        self.remote_pwd = remote_pwd
        self.extra = extra
        self.use_prudentia_lib = use_prudentia_lib

    def use_ssh_key(self):
        return self.remote_pwd is None

    def inventory(self):
        prudentia_python_interpreter = ' ansible_python_interpreter=' + sys.executable
        inv = '[' + self.hostname + ']\n' + self.ip
        if self.use_prudentia_lib:
            inv += prudentia_python_interpreter

        if 'local' not in self.hostname and '127.0.0.1' != self.ip and 'localhost' != self.ip:
            inv += '\n\n[localhost]'
            inv += '\nlocalhost ansible_connection=local' + prudentia_python_interpreter

        inv += '\n'
        return inv

    def __repr__(self):
        values = [self.playbook, self.hostname, self.ip, self.remote_user, '*****' if self.remote_pwd else '', xstr(self.extra)]
        return '%s -> (%s)' % (self.name, ', '.join(i for i in values if i and i.strip()))

    def to_json(self):
        json_obj = {
            'name': self.name,
            'playbook': self.playbook,
            'hostname': self.hostname,
            'ip': self.ip
        }
        if self.remote_user:
            json_obj.update({'remote_user': self.remote_user})
        if self.remote_pwd:
            json_obj.update({'remote_pwd': self.remote_pwd})
        if self.extra:
            json_obj.update({'extra': self.extra.to_json()})
        return json_obj

    @staticmethod
    def from_json(json_obj, extra_type):
        return Box(
            json_obj.get('name'),
            json_obj.get('playbook'),
            json_obj.get('hostname'),
            json_obj.get('ip'),
            json_obj.get('remote_user'),
            json_obj.get('remote_pwd'),
            extra_type.from_json(json_obj.get('extra')) if extra_type else None
        )



import logging
import os
from subprocess import PIPE, Popen
from threading import Thread
import sys


class BashCmd(object):
    def __init__(self, *cmd_args):
        self.cmd_args = cmd_args
        self.env = os.environ.copy()
        self.cwd = os.getcwd()
        self.show_output = True
        self.output_stdout = []
        self.output_stderr = []
        self.ON_POSIX = 'posix' in sys.builtin_module_names

    def set_env_var(self, var, value):
        if value:
            self.env[var] = value

    def set_cwd(self, cwd):
        self.cwd = cwd

    def set_show_output(self, b):
        self.show_output = b

    def print_output(self, out, err):
        try:
            for line in iter(out.readline, b''):
                if self.show_output:
                    print line.strip()
                self.output_stdout.append(line)
            for line in iter(err.readline, b''):
                print "ERR - ", line.strip()
                self.output_stderr.append(line)
        finally:
            out.close()
            err.close()

    def execute(self):
        try:
            p = Popen(args=self.cmd_args, bufsize=1, stdout=PIPE, stderr=PIPE, close_fds=self.ON_POSIX, env=self.env,
                      cwd=self.cwd)
            t = Thread(target=self.print_output, args=(p.stdout, p.stderr))
            t.daemon = True # thread dies with the program
            t.start()
            p.wait()
            self.stdout = "".join(self.output_stdout)
            self.stderr = "".join(self.output_stderr)
            self.returncode = p.returncode
        except Exception as e:
            logging.exception('Command not executed.')
            print("ERROR - Problem running {0}: {1}".format(self.cmd_args, e))

    def __repr__(self):
        return "{0}" \
               "\nReturn code: {1}" \
               "\nStd Output: {2}" \
               "\nStd Error: {3}".format(self.cmd_args, self.returncode, self.stdout, self.stderr)

    def is_ok(self):
        if not self.returncode:
            return True
        else:
            return False

    def output(self):
        return self.stdout


import sys
from datetime import datetime
from random import randint

from ansible.inventory import Inventory
from ansible.playbook import PlayBook
from ansible.runner import Runner
from ansible import callbacks, errors
import ansible.constants as C
from ansible.color import stringc

from prudentia.domain import Box


def run_playbook(playbook_file, inventory, remote_user=C.DEFAULT_REMOTE_USER,
                 remote_pass=C.DEFAULT_REMOTE_PASS, transport=C.DEFAULT_TRANSPORT, extra_vars=None, only_tags=None):
    stats = callbacks.AggregateStats()
    playbook_cb = callbacks.PlaybookCallbacks(verbose=True)
    runner_cb = callbacks.PlaybookRunnerCallbacks(stats, verbose=True)
    playbook = PlayBook(
        playbook=playbook_file,
        inventory=inventory,
        remote_user=remote_user,
        remote_pass=remote_pass,
        transport=transport,
        extra_vars=extra_vars,
        only_tags=only_tags,
        callbacks=playbook_cb,
        runner_callbacks=runner_cb,
        stats=stats
    )

    provision_success = False
    try:
        start = datetime.now()
        playbook.run()

        hosts = sorted(playbook.stats.processed.keys())
        print callbacks.banner("PLAY RECAP")
        playbook_cb.on_stats(playbook.stats)
        for h in hosts:
            t = playbook.stats.summarize(h)
            print "%s : %s %s %s %s\n" % (
                _hostcolor(h, t),
                _colorize('ok', t['ok'], 'green'),
                _colorize('changed', t['changed'], 'yellow'),
                _colorize('unreachable', t['unreachable'], 'red'),
                _colorize('failed', t['failures'], 'red'))
            if t['unreachable'] == 0 and t['failures'] == 0:
                provision_success = True

        print "Play run took {0} minutes\n".format((datetime.now() - start).seconds / 60)
    except errors.AnsibleError, e:
        print >> sys.stderr, "ERROR: %s" % e
    return provision_success


def _colorize(lead, num, color):
    """ Print 'lead' = 'num' in 'color' """
    if num != 0 and color is not None:
        return "%s%s%-15s" % (stringc(lead, color), stringc("=", color), stringc(str(num), color))
    else:
        return "%s=%-4s" % (lead, str(num))


def _hostcolor(host, stats, color=True):
    if color:
        if stats['failures'] != 0 or stats['unreachable'] != 0:
            return "%-37s" % stringc(host, 'red')
        elif stats['changed'] != 0:
            return "%-37s" % stringc(host, 'yellow')
        else:
            return "%-37s" % stringc(host, 'green')
    return "%-26s" % host


def run_modules(items):
    for item in items:
        print item['summary']
        success, result = run_module(item['module'])
        if not success:
            return False
    return True


def run_module(runner):
    module_success = True
    module_result = ''
    results = runner.run()
    if len(results['dark']):
        module_success = False
        print 'Host not contacted: %s' % results
    else:
        for (hostname, result) in results['contacted'].items():
            if 'failed' in result:
                module_success = False
                module_result = result['msg']
                print 'Run failed: %s' % result['msg']
            else:
                module_result = result
    return module_success, module_result


def generate_inventory(box):
    tmp_inventory = '/tmp/prudentia-inventory-' + str(randint(1, 999999))
    f = None
    try:
        f = open(tmp_inventory, 'w')
        f.write(box.inventory())
    except IOError, e:
        print e
    finally:
        f.close()
    return Inventory(tmp_inventory)


def local_inventory():
    return generate_inventory(Box('local', None, 'local', '127.0.0.1', use_prudentia_lib=True))


def create_user(box):
    user = box.remote_user
    if 'root' not in user:
        if 'jenkins' in user:
            user_home = '/var/lib/jenkins'
        else:
            user_home = '/home/' + user

        inventory = generate_inventory(box)
        run_modules([
            {
                'summary': 'Wait for SSH port to become open ...',
                'module': Runner(
                    pattern='localhost',
                    inventory=inventory,
                    module_name='wait_for',
                    module_args='host={0} port=22 delay=10 timeout=60'.format(box.ip))
            },
            {
                'summary': 'Creating group \'{0}\' ...'.format(user),
                'module': Runner(
                    pattern=box.name,
                    inventory=inventory,
                    remote_user='root',
                    module_name='group',
                    module_args='name={0} state=present'.format(user))
            },
            {
                'summary': 'Creating user \'{0}\' ...'.format(user),
                'module': Runner(
                    pattern=box.name,
                    inventory=inventory,
                    remote_user='root',
                    module_name='user',
                    module_args='name={0} home={1} state=present shell=/bin/bash generate_ssh_key=yes group={0} groups=sudo'.format(
                        user, user_home)
                )
            },
            {
                'summary': 'Copy authorized_keys from root ...',
                'module': Runner(
                    pattern=box.name,
                    inventory=inventory,
                    remote_user='root',
                    module_name='command',
                    module_args="cp /root/.ssh/authorized_keys {0}/.ssh/authorized_keys".format(user_home)
                )
            },
            {
                'summary': 'Set permission on authorized_keys ...',
                'module': Runner(
                    pattern=box.name,
                    inventory=inventory,
                    remote_user='root',
                    module_name='file',
                    module_args="path={0}/.ssh/authorized_keys mode=600 owner={1} group={1}".format(user_home, user)
                )
            },
            {
                'summary': 'Ensuring sudoers no pwd prompting ...',
                'module': Runner(
                    pattern=box.name,
                    inventory=inventory,
                    remote_user='root',
                    module_name='lineinfile',
                    module_args="dest=/etc/sudoers state=present regexp=%sudo line='%sudo ALL=(ALL:ALL) NOPASSWD:ALL' validate='visudo -cf %s'"
                )
            }
        ])



import logging
from os.path import dirname
import re
import os
from abc import ABCMeta, abstractmethod
from cmd import Cmd
import random

from ansible.callbacks import DefaultRunnerCallbacks, AggregateStats
from ansible.inventory import Inventory
from ansible.playbook import PlayBook
from ansible.playbook.play import Play
import ansible.constants as C

from domain import Environment
from utils.provisioning import run_playbook, generate_inventory
from utils.io import prudentia_python_dir


class SimpleCli(Cmd):
    provider = None  # Set by his children

    def cmdloop(self, intro=None):
        try:
            Cmd.cmdloop(self, intro)
        except Exception as e:
            logging.exception('Got a nasty error.')
            print '\nGot a nasty error: %s\n' % e

    def _get_box(self, box_name):
        b = self.provider.env.get(box_name)
        if not b:
            print 'The box \'%s\' you entered does not exists.\n\nAfter typing the command press Tab for box suggestions.\n' % box_name
            return None
        else:
            return b

    def complete_box_names(self, text, line, begidx, endidx):
        completions = ['']
        tokens = line.split(' ')
        action = tokens[0]
        box_name = tokens[1]
        if len(tokens) <= 2:
            # boxes completion
            if not text:
                completions = [b.name for b in self.provider.boxes()]
            else:
                completions = [b.name for b in self.provider.boxes() if b.name.startswith(text)]
        else:
            if action == 'provision':
                # tags completion
                if not text:
                    completions = self.provider.tags[box_name][:]
                else:
                    completions = [t for t in self.provider.tags[box_name] if t.startswith(text)]
                current_tags = tokens[2:]
                completions = [c for c in completions if c not in current_tags]
        return completions


    def help_register(self):
        print "Registers a new box.\n"

    def do_register(self, line):
        self.provider.register()


    def help_reconfigure(self):
        print "Reconfigures an existing box.\n"

    def complete_reconfigure(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_reconfigure(self, line):
        box = self._get_box(line)
        if box:
            self.provider.reconfigure(box)


    def help_provision(self):
        print "Starts and provisions the box, it accepts as optional argument a playbook tag.\n"

    def complete_provision(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_provision(self, line):
        tokens = line.split(' ')
        box = self._get_box(tokens[0])
        if box:
            self.provider.provision(box, *tokens[1:])


    def help_unregister(self):
        print "Unregisters an existing box.\n"

    def complete_unregister(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_unregister(self, line):
        box = self._get_box(line)
        if box:
            self.provider.unregister(box)


    def help_set(self):
        print "Sets the value of an environment variable. " \
              "During provisioning it will forcibly override the one defined in any playbook.\n"

    def do_set(self, line):
        tokens = line.split(' ')
        variable = tokens[0]
        value = tokens[1]
        self.provider.set_var(variable, value)


    def help_unset(self):
        print "Unsets an existing environment variable.\n"

    def do_unset(self, line):
        self.provider.unset_var(line)


    def help_list(self):
        print "Shows a list of current boxes.\n"

    def do_list(self, line):
        boxes = self.provider.boxes()
        if not len(boxes):
            print 'No box has been registered yet.\n'
        else:
            for b in boxes:
                print b


    def do_EOF(self, line):
        print "\n"
        return True

    def emptyline(self, *args, **kwargs):
        return ""


class SimpleProvider(object):
    __metaclass__ = ABCMeta

    box_name_pattern = re.compile('- hosts: (.*)')

    def __init__(self, name, general_type=None, box_extra_type=None):
        self.env = Environment(name, general_type, box_extra_type)
        self.extra_vars = {'prudentia_dir': prudentia_python_dir()}
        self.tags = {}
        self.load_tags()
        self.provisioned = False

    def boxes(self):
        return self.env.boxes.values()

    def set_var(self, var, value):
        self.extra_vars[var] = value
        print "\nSet \'{0}\' -> {1}\n".format(var, value)

    def unset_var(self, var):
        self.extra_vars.pop(var, None)
        print "\nUnset \'{0}\'\n".format(var)

    def add_box(self, box):
        self.env.add(box)
        self.load_tags(box)

    def load_tags(self, box=None):
        for b in ([box] if box else self.boxes()):
            if not os.path.exists(b.playbook):
                print 'WARNING: Box \'{0}\' points to a NON existing playbook. ' \
                      'Please `reconfigure` or `unregister` the box.\n'.format(b.name)
            else:
                playbook = PlayBook(
                    playbook=b.playbook,
                    inventory=Inventory([]),
                    callbacks=DefaultRunnerCallbacks(),
                    runner_callbacks=DefaultRunnerCallbacks(),
                    stats=AggregateStats(),
                    extra_vars=self.extra_vars
                )
                play = Play(playbook, playbook.playbook[0], dirname(b.playbook))
                (matched_tags, unmatched_tags) = play.compare_tags('')
                self.tags[b.name] = list(unmatched_tags)

    def remove_box(self, box):
        if box.name in self.tags:
            self.tags.pop(box.name)
        return self.env.remove(box)

    @abstractmethod
    def register(self):
        pass

    @abstractmethod
    def reconfigure(self, box):
        pass

    def unregister(self, box):
        self.remove_box(box)
        print "\nBox %s removed.\n" % box.name

    def fetch_box_hostname(self, playbook):
        with open(playbook, 'r') as f:
            hostname = None
            for i, line in enumerate(f):
                if i == 1:  # 2nd line contains the host box_name
                    match = self.box_name_pattern.match(line)
                    hostname = match.group(1)
                elif i > 1:
                    break
        return hostname

    def suggest_name(self, hostname):
        if hostname not in self.env.boxes:
            return hostname
        else:
            return hostname + '-' + str(random.randint(0, 100))

    def provision(self, box, *tags):
        remote_user = C.DEFAULT_REMOTE_USER
        if box.remote_user:
            remote_user = box.remote_user

        remote_pwd = C.DEFAULT_REMOTE_PASS
        transport = C.DEFAULT_TRANSPORT
        if not box.use_ssh_key():
            remote_pwd = box.remote_pwd
            transport = 'paramiko'

        only_tags = None
        if tags is not ():
            only_tags = tags

        self.provisioned = run_playbook(
            playbook_file=box.playbook,
            inventory=generate_inventory(box),
            remote_user=remote_user,
            remote_pass=remote_pwd,
            transport=transport,
            extra_vars=self.extra_vars,
            only_tags=only_tags
        )



from abc import abstractmethod, ABCMeta

from simple import SimpleCli, SimpleProvider


class FactoryCli(SimpleCli):
    def help_create(self):
        print "Creates the box.\n"

    def complete_create(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_create(self, line):
        box = self._get_box(line)
        if box:
            self.provider.create(box)


    def help_start(self):
        print "Starts the box.\n"

    def complete_start(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_start(self, line):
        box = self._get_box(line)
        if box:
            self.provider.start(box)


    def help_restart(self):
        print "Restarts the box.\n"

    def complete_restart(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_restart(self, line):
        box = self._get_box(line)
        if box:
            self.provider.stop(box)
            self.provider.start(box)


    def help_phoenix(self):
        print "Regenerates a box: destroy -> create -> provision.\n"

    def complete_phoenix(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_phoenix(self, line):
        box = self._get_box(line)
        if box:
            self.provider.rebuild(box)
            self.provider.provision(box)


    def help_stop(self):
        print "Stops the box.\n"

    def complete_stop(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_stop(self, line):
        box = self._get_box(line)
        if box:
            self.provider.stop(box)


    def help_destroy(self):
        print "Destroys the box.\n"

    def complete_destroy(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_destroy(self, line):
        box = self._get_box(line)
        if box:
            self.provider.destroy(box)

    def help_status(self):
        print "Check the status of the box.\n"

    def complete_status(self, text, line, begidx, endidx):
        return self.complete_box_names(text, line, begidx, endidx)

    def do_status(self, line):
        box = self._get_box(line)
        if box:
            self.provider.status(box)


class FactoryProvider(SimpleProvider):
    __metaclass__ = ABCMeta

    def add_box(self, box):
        self.create(box)
        super(FactoryProvider, self).add_box(box)

    @abstractmethod
    def create(self, box):
        pass

    @abstractmethod
    def start(self, box):
        pass

    @abstractmethod
    def stop(self, box):
        pass

    def remove_box(self, box):
        box = super(FactoryProvider, self).remove_box(box)
        self.destroy(box)
        return box

    @abstractmethod
    def destroy(self, box):
        pass

    def rebuild(self, box):
        self.stop(box)
        self.destroy(box)
        self.create(box)
        self.start(box)

    @abstractmethod
    def status(self, box):
        pass



import logging
from os import path
import re

from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader

from domain import Box
from domain import Environment as PrudentiaEnv
from factory import FactoryProvider, FactoryCli
from simple import SimpleProvider
from utils.bash import BashCmd
from utils.io import input_value, input_yes_no, input_path


class VagrantCli(FactoryCli):
    def __init__(self):
        FactoryCli.__init__(self)
        self.prompt = '(Prudentia > Vagrant) '
        self.provider = VagrantProvider()


class VagrantProvider(FactoryProvider):
    NAME = 'vagrant'
    VAGRANT_FILE_NAME = 'Vagrantfile'
    ENV_DIR = path.join(PrudentiaEnv.DEFAULT_ENVS_PATH, NAME)
    CONF_FILE = path.join(ENV_DIR, VAGRANT_FILE_NAME)

    DEFAULT_USER = 'vagrant'
    DEFAULT_PWD = 'vagrant'

    def __init__(self):
        super(VagrantProvider, self).__init__(self.NAME, box_extra_type=VagrantExt)
        this_path = path.dirname(path.realpath(__file__))
        self.template_env = Environment(loader=FileSystemLoader(this_path), auto_reload=True)
        # BashCmd(path.join(this_path, 'install_vagrant.sh')).execute()

    def register(self):
        try:
            playbook = input_path('absolute playbook path')
            hostname = self.fetch_box_hostname(playbook)
            name = input_value('box name', self.suggest_name(hostname))
            ip = input_value('internal IP')

            ext = VagrantExt()
            mem = input_value('amount of RAM in GB', 1)
            ext.set_mem(mem * 1024)

            ext.set_shares(self._input_shares())

            box = Box(name, playbook, hostname, ip, self.DEFAULT_USER, self.DEFAULT_PWD, ext)
            self.add_box(box)
            print "\nBox %s added." % box
        except Exception as e:
            logging.exception('Box not added.')
            print '\nError: %s\n' % e

    def add_box(self, box):
        SimpleProvider.add_box(self, box)
        self.create(box)

    def remove_box(self, box):
        b = super(VagrantProvider, self).remove_box(box)
        self._generate_vagrant_file()
        return b

    def reconfigure(self, previous_box):
        try:
            self.remove_box(previous_box)

            playbook = input_path('absolute playbook path', previous_box.playbook)
            hostname = self.fetch_box_hostname(playbook)
            ip = input_value('internal IP', previous_box.ip)

            ext = VagrantExt()
            mem = input_value('amount of RAM in GB', previous_box.extra.mem / 1024)
            ext.set_mem(mem * 1024)

            ext.set_shares(self._input_shares())

            box = Box(previous_box.name, playbook, hostname, ip, self.DEFAULT_USER, self.DEFAULT_PWD, ext)
            self.add_box(box)
            print "\nBox %s reconfigured." % box
        except Exception as e:
            logging.exception('Box not reconfigured.')
            print '\nError: %s\n' % e

    def _input_shares(self):
        shares = []
        loop = True
        while loop:
            if input_yes_no('share a folder'):
                src = input_path('directory on the HOST machine', is_file=False)
                if not path.exists(src):
                    raise ValueError("Directory '%s' on the HOST machine doesn't exists." % src)
                dst = input_value('directory on the GUEST machine')
                shares.append((src, dst))
            else:
                loop = False
        return shares

    def _generate_vagrant_file(self):
        env = self.template_env
        template_name = self.VAGRANT_FILE_NAME + '.j2'
        template = env.get_template(template_name)
        template.stream({
            'boxes': self.boxes()
        }).dump(self.CONF_FILE)

    def create(self, box):
        self._generate_vagrant_file()
        self.start(box)

    def start(self, box):
        self._action(action="up", action_args=("--no-provision", box.name))

    def stop(self, box):
        self._action(action="halt", action_args=(box.name,))

    def destroy(self, box):
        if input_yes_no('destroy the instance \'{0}\''.format(box.name)):
            self._action(action="destroy", action_args=("-f", box.name))

    def status(self, box):
        output = self._action(action="status", action_args=(box.name,), output=False)
        pattern = '.*' + box.name + '\s*(.*?) \(virtualbox\).*'
        match = re.match(pattern, output, re.DOTALL)
        print match.group(1)

    def _action(self, **kwargs):
        if 'action_args' not in kwargs.keys():
            cmd = BashCmd("vagrant", kwargs['action'])
        else:
            cmd = BashCmd("vagrant", kwargs['action'], *kwargs['action_args'])

        cmd.set_cwd(self.ENV_DIR)
        if 'output' in kwargs.keys():
            cmd.set_show_output(kwargs['output'])
        if 'tags' in kwargs.keys():
            cmd.set_env_var("TAGS", kwargs['tags'])

        # for debugging
        # cmd.set_env_var("VAGRANT_LOG", "INFO")

        cmd.execute()
        if not cmd.is_ok():
            print "ERROR while running: {0}".format(cmd.cmd_args)
        else:
            return cmd.output()


class VagrantExt(object):
    mem = None
    shares = None

    def set_mem(self, mem):
        self.mem = mem

    def set_shares(self, shares):
        self.shares = shares

    def __repr__(self):
        return 'VagrantExt[mem: %s, shares: %s]' % (self.mem, self.shares)

    def to_json(self):
        return {'mem': self.mem, 'shares': self.shares}

    @staticmethod
    def from_json(json):
        e = VagrantExt()
        e.set_mem(json['mem'])
        e.set_shares(json['shares'])
        return e



import os
from getpass import getpass


def prudentia_python_dir():
    cwd = os.path.realpath(__file__)
    components = cwd.split(os.sep)
    last_prudentia_index = len(components) - components[::-1].index("prudentia")
    return str.join(os.sep, components[:last_prudentia_index])


def xstr(s):
    return '' if s is None else str(s)


def first_time_input():
    if first_time_input.show:
        print '\nPlease enter values for the following settings, ' \
              'press \'Enter\' to accept the default value if given in brackets \'[ ]\'.\n'
        first_time_input.show = False
first_time_input.show = True


def input_value(topic, default_value=None, default_description=None, mandatory=True, hidden=False):
    first_time_input()
    default = default_description if default_description else default_value
    if default:
        input_msg = 'Specify the %s [default: %s]: ' % (topic, default)
    else:
        input_msg = 'Specify the %s: ' % topic
    if not hidden:
        answer = raw_input(input_msg)
    else:
        answer = getpass(input_msg)
    answer = answer.strip()
    if not len(answer):
        if default_value:
            answer = default_value
        else:
            if mandatory:
                raise ValueError('You must enter a valid %s.' % topic)
            else:
                answer = None
    else:
        if default_value and type(default_value) == int:
            answer = int(answer)
    return answer


def input_path(topic, default_value=None, default_description=None, mandatory=True, hidden=False, is_file=True):
    path = input_value(topic, default_value, default_description, mandatory, hidden)
    if not os.path.exists(path):
        raise ValueError('The %s you entered does NOT exist.' % topic)
    elif not os.path.isabs(path):
        raise ValueError('The %s you entered is NOT absolute.' % topic)
    elif is_file and not os.path.isfile(path):
        raise ValueError('The %s you entered is NOT a file.' % topic)
    elif not is_file and not os.path.isdir(path):
        raise ValueError('The %s you entered is NOT a directory.' % topic)
    else:
        return path


def input_yes_no(topic, default='n'):
    first_time_input()
    input_msg = 'Do you want to %s? [default: %s]: ' % (topic, default.upper())
    answer = raw_input(input_msg).strip()
    if not len(answer):
        answer = default
    if answer.lower() in ('y', 'yes'):
        return True
    else:
        return False



import os
from os import path

# Setting Ansible config file environment variable as first thing
os.environ['ANSIBLE_CONFIG'] = path.join(path.dirname(path.realpath(__file__)), 'ansible.cfg')

from cmd import Cmd

from digital_ocean import DigitalOceanCli
from local import LocalCli
from ssh import SshCli
from vagrant import VagrantCli

Environments = {
    'local': LocalCli,
    'ssh': SshCli,
    'vagrant': VagrantCli,
    'digital-ocean': DigitalOceanCli
}


class CLI(Cmd):
    parent_loop = False
    env_cli = None

    def __init__(self, *args, **kwargs):
        Cmd.__init__(self, *args, **kwargs)
        self.prompt = '(Prudentia) '

    def cmdloop(self, *args, **kwargs):
        self.parent_loop = True
        print '\nTo start: `use` one of the available providers: %s\n' % ', '.join(
            str(p) for p in Environments.keys())
        return Cmd.cmdloop(self, *args, **kwargs)

    def complete_use(self, text, line, begidx, endidx):
        if not text:
            return Environments.keys()
        else:
            return [e for e in Environments.keys() if e.startswith(text)]

    def do_use(self, env, *args):
        result = False
        if env in Environments.keys():
            self.env_cli = Environments[env]()
            if args:
                cmd = ' '.join(args)
                print "Executing: '{0}'\n".format(cmd)
                self.env_cli.onecmd(cmd)
            else:
                self.env_cli.cmdloop()

            result = self.env_cli.provider.provisioned
        else:
            print "Provider '{0}' NOT found.".format(env)

        # If this function was called inside a cmd loop the return values indicates whether execution will be terminated
        # returning False will cause interpretation to continue.
        # Otherwise the return value is the result of the 'provision' action.
        return False if self.parent_loop else result

    def do_EOF(self, line):
        print "\n\nBye!"
        return True

    def emptyline(self, *args, **kwargs):
        return ""



# !/usr/bin/env python

import sys

from prudentia import __version__, __author__

from codecs import open

try:
    from setuptools import setup
except ImportError:
    print "Prudentia needs setuptools in order to build. " \
          "Install it using your package manager (usually python-setuptools) or via pip (pip install setuptools)."
    sys.exit(1)

with open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()
with open('HISTORY.rst', 'r', 'utf-8') as f:
    history = f.read()

setup(
    name='prudentia',
    version=__version__,
    description='Continuous Deployment toolkit.',
    author=__author__,
    author_email='tiziano@startersquad.com',
    long_description=readme + '\n\n' + history,
    url='https://github.com/StarterSquad/prudentia',
    license='MIT',
    install_requires=['ansible', 'dopy', 'boto'],
    packages=['prudentia', 'prudentia.utils'],
    include_package_data=True,
    scripts=['bin/prudentia'],
    platforms='Posix; MacOS X;'
)



import logging

import ansible.constants as C

from domain import Box
from simple import SimpleProvider, SimpleCli
from utils.io import input_value, input_path


class SshCli(SimpleCli):
    def __init__(self):
        SimpleCli.__init__(self)
        self.prompt = '(Prudentia > Ssh) '
        self.provider = SshProvider()


class SshProvider(SimpleProvider):
    NAME = 'ssh'

    def __init__(self):
        super(SshProvider, self).__init__(self.NAME)

    def register(self):
        try:
            playbook = input_path('absolute playbook path')
            hostname = self.fetch_box_hostname(playbook)
            name = input_value('box name', self.suggest_name(hostname))
            ip = input_value('ip or hostname of the instance')
            user = input_value('remote user', C.active_user)
            pwd = input_value('password for the remote user', default_description='ssh key', mandatory=False, hidden=True)

            box = Box(name, playbook, hostname, ip, user, pwd)
            self.add_box(box)
            print "\nBox %s added." % box
        except Exception as e:
            logging.exception('Box not added.')
            print '\nError: %s\n' % e

    def reconfigure(self, previous_box):
        try:
            self.remove_box(previous_box)

            playbook = input_path('absolute playbook path', previous_box.playbook)
            hostname = self.fetch_box_hostname(playbook)
            ip = input_value('ip or hostname of the instance', previous_box.ip)
            user = input_value('remote user', previous_box.remote_user)
            if previous_box.remote_pwd:
                pwd = input_value('password for the remote user', default_value=previous_box.remote_pwd, default_description='*****', mandatory=False, hidden=True)
            else:
                pwd = input_value('password for the remote user', default_description='ssh key', mandatory=False, hidden=True)

            box = Box(previous_box.name, playbook, hostname, ip, user, pwd)
            self.add_box(box)
            print "\nBox %s reconfigured." % box
        except Exception as e:
            logging.exception('Box not reconfigured.')
            print '\nError: %s\n' % e
